@@ -349,14 +349,23 @@ fn deserialize_map_of_maps() {
     let map = hashmap_it(map);
     let map = hashmap_it(map);
     let map = hashmap_it(map);
+    let ok_map = Ok(map);
 
-    let deserialized = from_str(
-        "key[key][key][key][key][key][key][key][key][key][key][key]\
-        [key][key][key][key][key][key][key][key][key]=value",
-    )
-    .unwrap();
+    assert_eq!(
+        from_str(
+            "key[key][key][key][key][key][key][key][key][key][key][key]\
+             [key][key][key][key][key][key][key][key][key]=value",
+        ),
+        ok_map
+    );
 
-    assert_eq!(map, deserialized);
+    assert_eq!(
+        from_str(
+            "key%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D\
+             %5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D%5Bkey%5D=value",
+        ),
+        ok_map
+    );
 }
 
 #[test]
@@ -922,6 +931,17 @@ fn deserialize_bench_multilevel() {
     );
 }
 
+#[derive(Deserialize, PartialEq)]
+struct SeqStruct {
+    value: Vec<i32>,
+}
+
+impl std::fmt::Debug for SeqStruct {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        f.debug_struct("Long sequence of values").finish()
+    }
+}
+
 #[test]
 fn deserialize_bench_seq() {
     let mut ordered = String::new();
@@ -938,16 +958,112 @@ fn deserialize_bench_seq() {
     }
     reverse.remove(reverse.len() - 1);
 
-    assert_eq!(from_str(&ordered), Ok(p!(res_ordered.clone())));
-    assert_eq!(from_str(&reverse), Ok(p!(res_ordered)));
+    assert_eq!(
+        from_str(&ordered),
+        Ok(SeqStruct {
+            value: res_ordered.clone()
+        })
+    );
+    assert_eq!(from_str(&reverse), Ok(SeqStruct { value: res_ordered }));
 }
 
 #[test]
-fn deserialize_invalid() {
-    // from_str::<HashMap<String, Vec<i32>>>("x[3]=22&&x[2]")
+fn deserialize_percent_decoded_brackets() {
+    // We accept percent-decoded brackets either if they used as both the opening and closing brackets
+    // Or if they used with normal brackets together in any combination
+    // It may or may not make sense to accept a combination, please open an issue if you know of
+    // a situation that this may matter
+    assert_eq!(
+        from_str("value%5B%5D=22&value%5B1%5D=23&value[2%5D=24&value%5B3]=25&value[4]=26"),
+        Ok(p!(vec![22, 23, 24, 25, 26]))
+    );
+
+    assert_eq!(
+        from_str(
+            "value%5Bvalue%5D%5B%5D=22&value%5Bvalue%5D%5B1%5D=23&\
+            value[value][2%5D=24&value%5Bvalue]%5B3]=25&value[value][4]=26"
+        ),
+        Ok(p!(p!(vec![22, 23, 24, 25, 26])))
+    );
 }
 
 #[test]
-fn deserialize_to_unit() {
-    // from_str::<HashMap<String, ()>>("x[3]=22&x[2]=22") also for struct fields
+fn deserialize_invalid_percent_decoded() {
+    // We just ignore invalid %__ and we move on without decoding them, % with one trailing
+    // character is also accepted but we just don't parse it. For this reason %%5D is also
+    // accepted and will be decoded as `%[`. If you think it makes to do something else,
+    // maybe giving back an error, or checking if % comes with 2 trailing characters,
+    // please open an issue.
+    assert_eq!(
+        from_str("value[1%25]=21&value[2%%%]=22&value[3%PO]=23"),
+        Ok(p!(vec![21, 22, 23]))
+    );
+
+    let mut map = HashMap::new();
+    map.insert("%PO", 1);
+    map.insert("%%", 2);
+    map.insert("%%%", 3);
+    map.insert("%query", 4);
+
+    assert_eq!(from_str("%PO=1&%%=2&%%%=3&%query=4"), Ok(map.clone()));
+
+    assert_eq!(
+        from_str("value[%PO]=1&value[%%]=2&value[%%%]=3&value[%query]=4"),
+        Ok(p!(map.clone()))
+    );
+
+    assert_eq!(
+        from_str(
+            "value[value][%PO]=1&value[value][%%]=2&value[value][%%%]=3&value[value][%query]=4"
+        ),
+        Ok(p!(p!(map)))
+    );
+}
+
+#[test]
+fn deserialize_invalid_map_keys() {
+    assert!(from_str::<Primitive<HashMap<String, i32>>>("value[[]]=2").is_err());
+    assert!(from_str::<Primitive<Primitive<HashMap<String, i32>>>>("value[value]value=2").is_err());
+    assert!(from_str::<Primitive<HashMap<String, i32>>>("value[a=2").is_err());
+
+    // Key is [a
+    assert!(from_str::<Primitive<HashMap<String, i32>>>("value[[a]=2").is_ok());
+
+    // Key is value]
+    assert!(from_str::<HashMap<String, HashMap<String, i32>>>("value][a]=2").is_ok());
+
+    // The same is not valid for subkeys, not yet at least
+    assert!(from_str::<Primitive<HashMap<String, i32>>>("value[value]]=2").is_err());
+}
+
+#[test]
+fn deserialize_invalid_sequence() {
+    // We can either visit empty values in sequence as Option::None or unit
+    assert_eq!(
+        from_str("value=,1,,2,,,3,"),
+        Ok(p!(vec![
+            None,
+            Some(1),
+            None,
+            Some(2),
+            None,
+            None,
+            Some(3),
+            None
+        ]))
+    );
+    assert_eq!(
+        from_str("value=,1,2,,,3,"),
+        Ok(p!(((), 1, 2, (), (), 3, ())))
+    );
+
+    let mut map = HashMap::new();
+    map.insert("value", ((), 1, 2, (), (), 3, ()));
+    map.insert("value2", ((), 1, 2, (), (), 3, ()));
+    assert_eq!(from_str("value=,1,2,,,3,&value2=,1,2,,,3,"), Ok(map));
+
+    // But we can't ignore them
+    assert!(from_str::<Primitive<Vec<i32>>>("value=1,2,3,").is_err());
+    assert!(from_str::<Primitive<Vec<i32>>>("value=,1,2,3").is_err());
+    assert!(from_str::<Primitive<Vec<i32>>>("value=1,2,,3").is_err());
 }
